package Teo;

# Teorethical calculations on seeing

use Exporter;
use PDL;
use PDL::GSL::SF;

@ISA = ('Exporter');
@EXPORT = qw(OTFcirc OTFlatm OTFsatm DvK DK OTFr sum_r0);

use strict;

sub sum_r0 {
  my @r0s = @_;
  my $r0 = 0;
  my $k = (-5.0/3.0);
  my $i;
  
  foreach $i (@r0s) {
    $r0 += PDL::pow($i,$k);
  }
  return PDL::pow($r0,(1.0/$k));
}

sub DvK {
  my ($r,$r0,$L0) = @_;

  return 6.16 * PDL::pow($r0,-5.0/3.0) * ((3.0/5.0)*PDL::pow($L0/(2*3.141592),5.0/3.0) - PDL::pow($r*$L0/(4*3.141592),5.0/6.0)*gsl_sf_bessel_Knu(2*3.141592*$r/$L0,5.0/6.0)/0.940656);
}

sub DK {
  my ($r,$r0) = @_;

  return 6.88*PDL::pow($r/$r0,5.0/3.0);
}


sub OTFcirc {
  my ($R,$n,$D,$lambda) = @_;
  my ($OTF,$r);
  my $rho = $D/$lambda;

  $r = xvals($n+1)*$R/$n;
  $OTF = $r <= $rho;
  $OTF *= (2/3.141592) * (PDL::acos($r/$rho) - ($r/$rho) * sqrt(1 - ($r/$rho)*($r/$rho)));
}

sub OTFr {
  my ($R,$n) = @_;
  my $r;

  $r = xvals($n+1);
  $r *= $R/$n;

  return $r;
}

sub OTFlatm {
  my ($R,$n,$r0,$lambda,$L0) = @_;
  my ($D,$Dt,$r,$O);
  $r = xvals($n+1);
  $r *= $R/$n;
  $r *= $lambda;

  $D = zeroes($n+1);
  $Dt = $D->mslice([1,$n]);

  if (defined $L0) {
    $Dt .= DvK($r->mslice([1,$n]),$r0,$L0);
  }
  else {
    $Dt .= DK($r->mslice([1,$n]),$r0);
  }
  $O = exp(-0.5*$D);
  $O /= $O->at(0);
}

sub OTFsatm {
  my ($R,$n,$r0,$lambda,$L0,$diam) = @_;
  my ($D,$Dt,$r,$O);
  $r = xvals($n+1);
  $r *= $R/$n;
  $r *= $lambda;

  $D = zeroes($n+1);
  $Dt = $D->mslice([1,$n]);

  if (defined $L0) {
    $Dt .= DvK($r->mslice([1,$n]),$r0,$L0);
    $D -= $r*$r*(DvK($diam,$r0,$L0)/($diam*$diam));
  }
  else {
    $Dt .= DK($r->mslice([1,$n]),$r0);
    $D -= $r*$r*(DK($diam,$r0)/($diam*$diam));
  }
  $O = exp(-0.5*$D);
  $O /= $O->at(0);
}


1;
