package Utils;

# Utility functions

use Exporter;
use PDL;

@ISA = ('Exporter');
@EXPORT = qw(recenter shift2d rfcut maxpix hankel pi r2c);

use strict;

sub pi {4*PDL::atan(1);}

sub r2c {
  my ($im) = @_;
  my ($xn,$yn) = $im->dims;
  my ($r,$tmp);

  $r = zeroes $im->type,2,$xn,$yn;
  $tmp = $r->slice("(0),:,:");
  $tmp .= $im;
  return $r;
}

sub recenter {
  my $im = shift;
  my ($xc,$yc,$s);

  $s = sumover($im->clump(2));
  $xc = sumover(($im*xvals(($im->dims)[0]))->clump(2)) / $s;
  $yc = sumover(($im->xchg(0,1)*xvals(($im->dims)[1]))->clump(2)) / $s;
  return shift2d($im, int ( (($im->dims)[0] - 1)/2 - $xc ), 
		 int ( (($im->dims)[1] - 1)/2 - $yc ) );
}

sub shift2d {
  my ($im,$x,$y) = @_;

  return $im->rotate($x)->xchg(0,1)->rotate($y)->xchg(0,1);
}

sub rfcut {
  my ($im,$fmax) = @_;
  my $mask;

  $mask = rvals( float,$fmax/2,$fmax,{Centre=>[0,($fmax-1)/2]} ) ;
  $mask = $mask <= $fmax/2;
  $im = shift2d($im,0,int ( ($im->dims)[1] / 2 ) );
  return $mask * $im->mslice([0,$fmax/2-1],
			     [(($im->dims)[1] / 2)-$fmax/2, (($im->dims)[1] / 2)+$fmax/2-1]);
}

sub maxpix {
  my $in = shift;
  my $x = pdl 0;
  my $y = pdl 0;
  
  maximum_n_ind($in->maximum,$x);
  maximum_n_ind($in->xchg(0,1)->maximum,$y);
  return ($y->at,$x->at);
}

sub hankel {
  my ($dr,$g,$rho) = @_;
  my ($r,$n,$i,$G,$t,$maxr);

  $n = ($rho->dims)[0];
  $G = zeroes $n;
  $r = sequence(($g->dims)[0])*$dr;
  $maxr = max $r;
  for($i=0;$i<$n;$i++) {
    $t = $g*PDL::bessj0(2*pi()*$r*$rho->at($i))*$r;
    $G->set($i,2*pi()*$dr*($t->intover->at))
  }
  return $G;
}

1;
