
#include "gsl_fftw_wrapper.h"

#include <stdlib.h>
#include <gsl_complex_op.h>
#include <gsl_errno.h>
#include <gsl_matrix_op.h>

#define MAX_PLANS 100

static fftwnd_plan planf[MAX_PLANS];
static fftwnd_plan planb[MAX_PLANS];
static int n[MAX_PLANS];
static int m[MAX_PLANS];
static int r[MAX_PLANS];

void gsl_fftw_init(void) {
  int i;
  
  for(i=0;i<MAX_PLANS;i++) r[i]=0;
}

int gsl_fftw_init_matrix_complex(gsl_matrix_complex *mat, gsl_fftw_plan *newplan, int keep) {
  int i;

  for(i=0;i<MAX_PLANS;i++) 
    if ( (r[i]>0) && (n[i]==mat->size1) && (m[i]==mat->size2) ) {
      r[i]++;
      newplan->matrix=mat;
      newplan->planf=&planf[i];
      newplan->planb=&planb[i];
      newplan->index=i;
      return GSL_SUCCESS;
    }
  for(i=0;i<MAX_PLANS;i++) 
    if (r[i]==0) {
      r[i]++;
      n[i]=mat->size1;
      m[i]=mat->size2;
      if (keep==GSL_FFTW_DESTROY) {
	planf[i]=fftw2d_create_plan_specific(n[i],m[i],FFTW_FORWARD,
						   FFTW_ESTIMATE | FFTW_IN_PLACE | FFTW_USE_WISDOM,
						   (FFTW_COMPLEX *) mat->data,1,NULL,1); /*B*/
	planb[i]=fftw2d_create_plan_specific(n[i],m[i],FFTW_BACKWARD,
						   FFTW_ESTIMATE | FFTW_IN_PLACE | FFTW_USE_WISDOM,
						   (FFTW_COMPLEX *) mat->data,1,NULL,1); /*B*/
      } 
      else {
	planf[i]=fftw2d_create_plan(n[i],m[i],FFTW_FORWARD,
					  FFTW_ESTIMATE | FFTW_IN_PLACE | FFTW_USE_WISDOM); /*B*/
	planb[i]=fftw2d_create_plan(n[i],m[i],FFTW_BACKWARD,
					  FFTW_ESTIMATE | FFTW_IN_PLACE | FFTW_USE_WISDOM); /*B*/
      }
      newplan->planf=&planf[i];
      newplan->planb=&planb[i];
      newplan->matrix=mat;
      newplan->index=i;
      return GSL_SUCCESS;
    }
  GSL_ERROR("No more free plans!",GSL_ENOMEM);
}

int gsl_fftw_matrix_complex_forward(gsl_fftw_plan *plan) {
  fftwnd(*(plan->planf),1,(FFTW_COMPLEX *) plan->matrix->data, 1,0,NULL,1,0);
  return GSL_SUCCESS;
}

int gsl_fftw_matrix_complex_backward(gsl_fftw_plan *plan) {
  fftwnd(*(plan->planb),1,(FFTW_COMPLEX *) plan->matrix->data, 1,0,NULL,1,0);
  return GSL_SUCCESS;
}

int gsl_fftw_matrix_complex_inverse(gsl_fftw_plan *plan) {

  fftwnd(*(plan->planb),1,(FFTW_COMPLEX *) plan->matrix->data, 1,0,NULL,1,0);
  gsl_matrix_complex_mul_complex(plan->matrix,1.0 /( ((double) plan->matrix->size1)*plan->matrix->size2),0);
  return GSL_SUCCESS;
}

int gsl_fftw_free_matrix_complex(gsl_fftw_plan *plan, int keep) {
  if (keep==GSL_FFTW_DESTROY) {
    gsl_matrix_complex_free(plan->matrix);
  }
  if ( (--r[plan->index]) == 0) {
    fftwnd_destroy_plan(planf[plan->index]);
    fftwnd_destroy_plan(planb[plan->index]);
  }
  return GSL_SUCCESS;
}
