
BASE FUNCTION (gsl_matrix, sum) ( TYPE(gsl_matrix) *a) {
  int i, max= a->size2 * a->size1;
  ATOMIC *p=a->data;
  BASE temp;

#if MULTIPLICITY>1
  GSL_REAL(temp)=0;
  GSL_IMAG(temp)=0;
#else
  temp=0;
#endif

  for(i=0;i<max;i++) {
#if MULTIPLICITY>1
    gsl_complex_add(temp,(*((BASE*) p)));
#else
    temp+= *((BASE*) p);
#endif
    p+=MULTIPLICITY;
  }
  return temp;
}

int FUNCTION (gsl_matrix, mul_scalar) ( TYPE(gsl_matrix) *a, double s) {
  int i, max= a->size2 * a->size1;
  ATOMIC *p=a->data;

  for(i=0;i<max;i++) {
#if MULTIPLICITY>1
    gsl_complex_multiply_reim(*((BASE*) p),s,0.0);
#else
    *((BASE*) p)*=s;
#endif
    p+=MULTIPLICITY;
  }
  return GSL_SUCCESS;
}

int FUNCTION (gsl_matrix, add_scalar) ( TYPE(gsl_matrix) *a, double s) {
  int i, max= a->size2 * a->size1;
  ATOMIC *p=a->data;

  for(i=0;i<max;i++) {
#if MULTIPLICITY>1
    fprintf(stderr,"BOING!\n");
    exit(1);
#else
    *((BASE*) p)+=s;
#endif
    p+=MULTIPLICITY;
  }
  return GSL_SUCCESS;
}

int FUNCTION (gsl_matrix, add) ( TYPE(gsl_matrix) *a, TYPE(gsl_matrix) *b) {
  int i, max= a->size2 * a->size1;
  ATOMIC *p=a->data;
  ATOMIC *t=b->data;
  
  if ( (a->size1 != b->size1) || (a->size2 != b->size2) ) {
    GSL_ERROR("Matrixes dimensions doesn't match",GSL_EINVAL);
  }
  
  for(i=0;i<max;i++) {
#if MULTIPLICITY>1
    gsl_complex_add( *((BASE*) p), *((BASE*) t) );
#else
    *p+=*t;
#endif
    p+=MULTIPLICITY;
    t+=MULTIPLICITY;
  }
  return GSL_SUCCESS;
}

int FUNCTION (gsl_matrix, mul) ( TYPE(gsl_matrix) *a, TYPE(gsl_matrix) *b) {
  int i, max= a->size2 * a->size1;
  ATOMIC *p=a->data;
  ATOMIC *t=b->data;
  
  if ( (a->size1 != b->size1) || (a->size2 != b->size2) ) {
    GSL_ERROR("Matrixes dimensions doesn't match",GSL_EINVAL);
  }

  for(i=0;i<max;i++) {
#if MULTIPLICITY>1
  gsl_complex_multiply( *((BASE*) p), *((BASE*) t) );
#else
  *p*=*t;
#endif
    p+=MULTIPLICITY;
    t+=MULTIPLICITY;
  }
  return GSL_SUCCESS;
}

int FUNCTION (gsl_matrix, sub) ( TYPE(gsl_matrix) *a, TYPE(gsl_matrix) *b) {
  int i, max= a->size2 * a->size1;
  ATOMIC *p=a->data;
  ATOMIC *t=b->data;
  
  if ( (a->size1 != b->size1) || (a->size2 != b->size2) ) {
    GSL_ERROR("Matrixes dimensions doesn't match",GSL_EINVAL);
  }
  
  for(i=0;i<max;i++) {
#if MULTIPLICITY>1
    gsl_complex_sub( *((BASE*) p), *((BASE*) t));
#else
    *p-=*t;
#endif
    p+=MULTIPLICITY;
    t+=MULTIPLICITY;
  }
  return GSL_SUCCESS;
}

int FUNCTION (gsl_matrix, div) ( TYPE(gsl_matrix) *a, TYPE(gsl_matrix) *b) {
  int i, max= a->size2 * a->size1;
  ATOMIC *p=a->data;
  ATOMIC *t=b->data;
  
  if ( (a->size1 != b->size1) || (a->size2 != b->size2) ) {
    GSL_ERROR("Matrixes dimensions doesn't match",GSL_EINVAL);
  }
  
  for(i=0;i<max;i++) {
#if MULTIPLICITY>1
    gsl_complex_divide( *((BASE*) p),*((BASE*) t) );
#else
    *p/=*t;
#endif
    p+=MULTIPLICITY;
    t+=MULTIPLICITY;
  }
  return GSL_SUCCESS;
}

int FUNCTION (gsl_matrix, clone) ( TYPE(gsl_matrix) *a, TYPE(gsl_matrix) **n){
  *n= FUNCTION(gsl_matrix, alloc) (a->size1, a->size2);
  
  memcpy((*n)->data,a->data,a->size1*a->size2*sizeof(ATOMIC)*MULTIPLICITY);
}

#if MULTIPLICITY>1
int FUNCTION (gsl_matrix, mul_complex) ( TYPE(gsl_matrix) *a, double re, double im){
  int i, max= a->size2 * a->size1;
  ATOMIC *p=a->data;
  
  for(i=0;i<max;i++) {
    gsl_complex_multiply_reim( *((BASE*) p), re, im);
    p+=MULTIPLICITY;
  }
  return GSL_SUCCESS;  
}

int FUNCTION (gsl_matrix, keep_real) ( TYPE(gsl_matrix) *a, gsl_matrix **b) {
  int i, max= a->size2 * a->size1;
  double tempo, *t;

  ATOMIC *p=a->data;
  *b= (gsl_matrix *) a;
  t=(*b)->data;

  for(i=0;i<max;i++) {
    tempo=GSL_REAL(*((BASE*) p));
    *t=tempo;
    p+=MULTIPLICITY;
    t++;
  }
  (*b)->data=(double *)realloc((*b)->data,(*b)->size1*(*b)->size2*sizeof(double));
  return GSL_SUCCESS;  
}

int FUNCTION (gsl_matrix, keep_mod) ( TYPE(gsl_matrix) *a, gsl_matrix **b) {
  int i, max= a->size2 * a->size1;
  double re,im,tempo, *t;

  ATOMIC *p=a->data;
  *b= (gsl_matrix *) a;
  t=(*b)->data;

  for(i=0;i<max;i++) {
    re=GSL_REAL(*((BASE*) p));
    im=GSL_IMAG(*((BASE*) p));
    tempo=sqrt(re*re+im*im);
    *t=tempo;
    p+=MULTIPLICITY;
    t++;
  }
  (*b)->data=(double *)realloc((*b)->data,(*b)->size1*(*b)->size2*sizeof(double));
  return GSL_SUCCESS;  
}

int FUNCTION (gsl_matrix, keep_mod2) ( TYPE(gsl_matrix) *a, gsl_matrix **b) {
  int i, max= a->size2 * a->size1;
  double re,im,tempo, *t;

  ATOMIC *p=a->data;
  *b= (gsl_matrix *) a;
  t=(*b)->data;

  for(i=0;i<max;i++) {
    re=GSL_REAL(*((BASE*) p));
    im=GSL_IMAG(*((BASE*) p));
    tempo=re*re+im*im;
    *t=tempo;
    p+=MULTIPLICITY;
    t++;
  }
  (*b)->data=(double *)realloc((*b)->data,(*b)->size1*(*b)->size2*sizeof(double));
  return GSL_SUCCESS;  
}

int FUNCTION (gsl_matrix, do_autocorr) ( TYPE(gsl_matrix) *a) {
  int i, max= a->size2 * a->size1;
  ATOMIC *p=a->data;
  
  for(i=0;i<max;i++) {
    double ti,tr;

    tr=GSL_REAL(*((BASE*) p));
    ti=GSL_IMAG(*((BASE*) p));
    gsl_complex_multiply_reim( *((BASE*) p), tr, ti);
    p+=MULTIPLICITY;
  }
  return GSL_SUCCESS;  
}

int FUNCTION (gsl_matrix, mod) ( TYPE(gsl_matrix) *a, gsl_matrix *b) {
  int i, max= a->size2 * a->size1;
  double tempo, *t,re ,im;

  ATOMIC *p=a->data;
  t=b->data;

  for(i=0;i<max;i++) {
    re=GSL_REAL(*((BASE*) p));
    im=GSL_IMAG(*((BASE*) p));
    *t=sqrt(re*re+im*im);
    p+=MULTIPLICITY;
    t++;
  }
  return GSL_SUCCESS;  
}

int FUNCTION (gsl_matrix, mod2) ( TYPE(gsl_matrix) *a, gsl_matrix *b) {
  int i, max= a->size2 * a->size1;
  double tempo, *t,re, im;

  ATOMIC *p=a->data;
  t=b->data;

  for(i=0;i<max;i++) {
    re=GSL_REAL(*((BASE*) p));
    im=GSL_IMAG(*((BASE*) p));
    *t=re*re+im*im;
    p+=MULTIPLICITY;
    t++;
  }
  return GSL_SUCCESS;  
}

int FUNCTION (gsl_matrix, phase) ( TYPE(gsl_matrix) *a, gsl_matrix *b) {
  int i, max= a->size2 * a->size1;
  double re,im, *t;

  ATOMIC *p=a->data;
  t=b->data;

  for(i=0;i<max;i++) {
    re=GSL_REAL(*((BASE*) p));
    im=GSL_IMAG(*((BASE*) p));
    *t=atan2(im,re);
    p+=MULTIPLICITY;
    t++;
  }
  return GSL_SUCCESS;  
}

int FUNCTION (gsl_matrix, set_real) ( TYPE(gsl_matrix) *a, gsl_matrix *b) {
  int i, max= a->size2 * a->size1;
  double tempo, *t;

  ATOMIC *p=a->data;
  t=b->data;

  for(i=0;i<max;i++) {
    GSL_REAL(*((BASE*) p))=*t;
    GSL_IMAG(*((BASE*) p))=0;
    p+=MULTIPLICITY;
    t++;
  }
  return GSL_SUCCESS;  
}

#else

int FUNCTION (gsl_matrix, rescale) (TYPE(gsl_matrix) *m, int rescale) {
  int i,j,ii,jj,rm,rn;
  BASE tot=rescale*rescale;

  rm=m->size1/rescale;
  rn=m->size2/rescale;
  for (i=0;i<(m->size1/rescale);i+=rescale)
    for (j=0;j<(m->size2/rescale);j+=rescale) {
      BASE t=0.0;

      for(ii=0;ii<rescale;ii++)
	for(jj=0;jj<rescale;jj++)
	  t += FUNCTION (gsl_matrix, get) (m,i+ii,j+jj);
      m->data[(i/rescale)*rn+(j/rescale)]=t/tot;
    }
  m->size1=rm;
  m->size2=rn;
  m->data=(BASE *)realloc(m->data,m->size1*m->size2*sizeof(double));
  return 1;
}

int FUNCTION (gsl_matrix, minmax) ( TYPE(gsl_matrix) *m, double *min, double *max, int *i_min, int *j_min,
				    int *i_max, int *j_max) {
  int i,j, maxi= m->size1, maxj=m->size2;
  ATOMIC *p=m->data;
  ATOMIC Max,Min;

  Max=*p;
  Min=*p;
  *i_max=*j_max=*i_min=*j_max=0;
  
  for(i=0;i<maxi;i++) 
    for (j=0;j<maxj;j++){
      ATOMIC t=*p;
      
      if (t>Max) {
	*i_max=i;
	*j_max=j;
	Max=t;
      }
      if (t<Min) {
	*i_min=i;
	*j_min=j;
	Min=t;
      }      
      p+=MULTIPLICITY;
    }
  *min=Min;
  *max=Max;
  return GSL_SUCCESS;
}

int FUNCTION (gsl_matrix, centroid) ( TYPE(gsl_matrix) *m, int *si, int *sj) {
  int i,j, maxi= m->size1, maxj=m->size2;
  ATOMIC *p=m->data;
  double dsi=0.0,dsj=0.0,sum=0.0;
  
  for(i=0;i<maxi;i++) 
    for (j=0;j<maxj;j++){
      ATOMIC t=*p;
      
      sum+=t;
      dsi+=t*i;
      dsj+=t*j;
      p+=MULTIPLICITY;
    }
  *si= (int) (dsi/sum);
  *sj= (int) (dsj/sum);
  return GSL_SUCCESS;
}

#endif


int FUNCTION (gsl_matrix, swap) ( TYPE(gsl_matrix) *a, int i0, int j0, int i1, int j1) {
  int i,j,j1_2;

  j1_2=j0+(j1-j0)/2;

  for (i=i0;i<i1;i++) 
    for (j=j0;j<j1_2;j++) {
      BASE t1,t2;

      t1= FUNCTION (gsl_matrix, get) (a,i,j);
      t2= FUNCTION (gsl_matrix, get) (a,i1-i+i0-1,j1-j+j0-1);
      FUNCTION (gsl_matrix, set) (a,i,j,t2);
      FUNCTION (gsl_matrix, set) (a,i1-i+i0-1,j1-j+j0-1,t1);
    }
}

int FUNCTION (gsl_matrix, rotate) ( TYPE(gsl_matrix) *a, int ri, int rj) {
  FUNCTION (gsl_matrix, swap) (a,0,0,ri,rj);
  FUNCTION (gsl_matrix, swap) (a,ri,0,a->size1,rj);
  FUNCTION (gsl_matrix, swap) (a,0,rj,ri,a->size2);
  FUNCTION (gsl_matrix, swap) (a,ri,rj,a->size1,a->size2);
  FUNCTION (gsl_matrix, swap) (a,0,0,a->size1,a->size2);
}


int FUNCTION (gsl_matrix, rot90) ( TYPE(gsl_matrix) *a) {
  int i,j;
  BASE t1,t2,t3,t4;

  for (i=0;i<(a->size1/2);i++) 
    for(j=0; j<(a->size2/2); j++) {
      t1=FUNCTION (gsl_matrix, get) (a,i,j);
      t2=FUNCTION (gsl_matrix, get) (a,a->size1-j-1,i);
      t3=FUNCTION (gsl_matrix, get) (a,a->size1-i-1,a->size2-j-1);
      t4=FUNCTION (gsl_matrix, get) (a,j,a->size2-i-1);
      FUNCTION (gsl_matrix, set) (a,i,j,t4);      
      FUNCTION (gsl_matrix, set) (a,a->size1-j-1,i,t1);      
      FUNCTION (gsl_matrix, set) (a,a->size1-i-1,a->size2-j-1,t2);      
      FUNCTION (gsl_matrix, set) (a,j,a->size2-i-1,t3);      
    } 
}

int FUNCTION (gsl_matrix, rot180) ( TYPE(gsl_matrix) *a) {
  int i,j;
  BASE t;

  for (i=0;i<(a->size1/2);i++) 
    for(j=0; j<(a->size2); j++) {
      t=FUNCTION (gsl_matrix, get) (a,i,j);
      FUNCTION (gsl_matrix, set) (a,i,j,
				  FUNCTION (gsl_matrix, get) (a,a->size1-i-1,a->size2-j-1) );      
      FUNCTION (gsl_matrix, set) (a,a->size1-i-1,a->size2-j-1,t);
    } 
}

int FUNCTION (gsl_matrix, rot270) ( TYPE(gsl_matrix) *a) {
  int i,j;
  BASE t1,t2,t3,t4;

  for (i=0;i<(a->size1/2);i++) 
    for(j=0; j<(a->size2/2); j++) {
      t1=FUNCTION (gsl_matrix, get) (a,i,j);
      t2=FUNCTION (gsl_matrix, get) (a,a->size1-j-1,i);
      t3=FUNCTION (gsl_matrix, get) (a,a->size1-i-1,a->size2-j-1);
      t4=FUNCTION (gsl_matrix, get) (a,j,a->size2-i-1);
      FUNCTION (gsl_matrix, set) (a,i,j,t2);      
      FUNCTION (gsl_matrix, set) (a,a->size1-j-1,i,t3);      
      FUNCTION (gsl_matrix, set) (a,a->size1-i-1,a->size2-j-1,t4);      
      FUNCTION (gsl_matrix, set) (a,j,a->size2-i-1,t1);      
    } 
}
